@@ -60,12 +60,21 @@ from .monkey_patching import CometModuleFinder
 from .offline import OfflineExperiment
 from .optimizer import Optimizer
 from .rpc import create_remote_call, get_remote_action_definition
-from .utils import Histogram, get_comet_version, merge_url, valid_ui_tabs
+from .utils import (
+    ConfusionMatrix,
+    Embedding,
+    Histogram,
+    get_comet_version,
+    merge_url,
+    valid_ui_tabs,
+)
 
 __author__ = "Gideon<Gideon@comet.ml>"
 __all__ = [
     "API",
     "APIExperiment",
+    "ConfusionMatrix",
+    "Embedding",
     "ExistingExperiment",
     "Experiment",
     "Histogram",
@@ -544,6 +553,112 @@ class Experiment(BaseExperiment):
         except Exception:
             LOGGER.error(SEND_NOTIFICATION_FAILED, exc_info=True)
 
+    def log_embedding(
+        self,
+        vectors,
+        labels,
+        image_data=None,
+        image_size=None,
+        image_preprocess_function=None,
+        image_transparent_color=None,
+        image_background_color_function=None,
+        title="Comet Embedding",
+        template_filename="template_projector_config.json",
+        group=None,
+    ):
+        """
+        Log a multi-dimensional dataset and metadata for viewing with
+        Comet's Embedding Projector (experimental).
+
+        Args:
+            vectors: the tensors to visualize in 3D
+            labels: labels for each tensor
+            image_data: (optional) list of arrays or Images
+            image_size: (optional, required if image_data is given) the size of each image
+            image_preprocess_function: (optional) if image_data is an
+                array, apply this function to each element first
+            image_transparent_color: a (red, green, blue) tuple
+            image_background_color_function: a function that takes an
+                index, and returns a (red, green, blue) color tuple
+            title: (optional) name of tensor
+            template_filename: (optional) name of template JSON file
+
+        See also: `Experiment._log_embedding_list()` and `comet_ml.Embedding`
+
+        Example:
+
+        ```python
+        from comet_ml import Experiment
+
+        import numpy as np
+        from keras.datasets import mnist
+
+        (x_train, y_train), (x_test, y_test) = mnist.load_data()
+
+        def label_to_color(index):
+            label = y_test[index]
+            if label == 0:
+                return (255, 0, 0)
+            elif label == 1:
+                return (0, 255, 0)
+            elif label == 2:
+                return (0, 0, 255)
+            elif label == 3:
+                return (255, 255, 0)
+            elif label == 4:
+                return (0, 255, 255)
+            elif label == 5:
+                return (128, 128, 0)
+            elif label == 6:
+                return (0, 128, 128)
+            elif label == 7:
+                return (128, 0, 128)
+            elif label == 8:
+                return (255, 0, 255)
+            elif label == 9:
+                return (255, 255, 255)
+
+        experiment = Experiment(project_name="projector-embedding")
+
+        experiment.log_embedding(
+            vectors=x_test,
+            labels=y_test,
+            image_data=x_test,
+            image_preprocess_function=lambda matrix: np.round(matrix/255,0) * 2,
+            image_transparent_color=(0, 0, 0),
+            image_size=(28, 28),
+            image_background_color_function=label_to_color,
+        )
+        ```
+        """
+        if not self.alive:
+            return None
+
+        LOGGER.warning(
+            "Logging embedding is experimental - the API and logged data are subject to change"
+        )
+
+        embedding = self._create_embedding(
+            vectors,
+            labels,
+            image_data,
+            image_size,
+            image_preprocess_function,
+            image_transparent_color,
+            image_background_color_function,
+            title,
+        )
+
+        if group is not None:
+            self._embedding_groups[group].append(embedding)
+            return embedding
+        else:
+            # Log the template:
+            template = {"embeddings": [embedding.to_json()]}
+            return self._log_asset_data(
+                template, template_filename, asset_type="embeddings"
+            )
+
 
 class ExistingExperiment(Experiment):
     """Existing Experiment allows you to report information to an
