@@ -17,6 +17,7 @@ from __future__ import print_function
 import io
 import logging
 import os
+import os.path
 
 import six
 from everett.manager import (
@@ -30,12 +31,23 @@ from everett.manager import (
     parse_bool,
 )
 
-from ._typing import Any
+from ._typing import Any, Optional
 
 try:  # everett version 1.0.1 or greater
     from everett.ext.inifile import ConfigIniEnv
-except ImportError:  # everett version 0.9
-    from everett.manager import ConfigIniEnv
+except ImportError:  # everett version 0.9 for Python 2
+    from everett.manager import ConfigIniEnv as UpstreamConfigIniEnv
+
+    class ConfigIniEnv(UpstreamConfigIniEnv):
+        """ Backport of everett ConfigIniEnv in Python 3 that save the path when found
+        """
+
+        def __init__(self, possible_paths):
+            super(ConfigIniEnv, self).__init__(possible_paths)
+
+            if self.cfg:
+                self.path = os.path.abspath(os.path.expanduser(possible_paths.strip()))
+
 
 LOGGER = logging.getLogger(__name__)
 
@@ -119,9 +131,9 @@ class Config(object):
             [  # User-defined overrides
                 ConfigOSEnv(),
                 ConfigEnvFileEnv(".env"),
-                ConfigIniEnv(
-                    [os.environ.get("COMET_INI"), "./.comet.config", "~/.comet.config"]
-                ),
+                ConfigIniEnv(os.environ.get("COMET_INI")),
+                ConfigIniEnv("./.comet.config"),
+                ConfigIniEnv("~/.comet.config"),
                 # Comet-defined overrides
                 self.backend_override,
             ],
@@ -291,6 +303,18 @@ class Config(object):
                     ini_file.write(six.u("%s = %s\n" % (setting, value)))
         print(" done!")
 
+    def get_config_origin(self, name):
+        # type: (str) -> Optional[str]
+        splitted = name.split(".")
+
+        for env in self.manager.envs:
+            value = env.get(splitted[-1], namespace=splitted[:-1])
+
+            if value != NO_VALUE:
+                return env
+
+        return None
+
 
 CONFIG_MAP = {
     "comet.disable_auto_logging": {"type": int, "default": 0},
@@ -334,7 +358,8 @@ CONFIG_MAP = {
     },
     "comet.timeout.http": {"type": int, "default": 10},
     "comet.timeout.api": {"type": int, "default": 10},
-    "comet.timeout.file_upload": {"type": int, "default": 180},
+    "comet.timeout.file_upload": {"type": int, "default": 900},
+    "comet.timeout.predictor": {"type": int, "default": 60},
     # HTTP Allow header
     "comet.allow_header.name": {"type": str},
     "comet.allow_header.value": {"type": str},
