@@ -21,7 +21,6 @@ information
 """
 
 import inspect
-import io
 import logging
 import os
 import platform
@@ -30,9 +29,11 @@ import sys
 
 import netifaces
 import requests
+import six
+from six import StringIO
 from six.moves.urllib.parse import urlparse
 
-from ._typing import Any, Callable, Dict, List, Optional, TypedDict, cast
+from ._typing import Any, Dict, List, Optional, Tuple
 from .connection import get_backend_address
 from .messages import SystemDetailsMessage
 from .utils import get_user
@@ -180,30 +181,116 @@ def get_caller_source_code():
             return None  # perhaps in ipython interactive
 
 
-def get_jupyter_source_code():
+def get_ipython_source_code():
+    # type: () -> str
     """
-    Get the Jupyter source code from the history. Assumes that this
+    Get the IPython source code from the history. Assumes that this
     method is run in a jupyter environment.
 
     Returns the command-history as a string that lead to this point.
     """
-
-    def in_format(n):
-        return "_i%s" % n
-
     import IPython
 
     ipy = IPython.get_ipython()
-    source = io.StringIO()
-    n = 1
-    while in_format(n) in ipy.ns_table["user_local"]:
-        source.write("# %%%% In [%s]:\n" % n)
-        source.write(ipy.ns_table["user_local"][in_format(n)])
-        source.write("\n\n")
-        n += 1
+    env = ipy.ns_table["user_local"]
+    source = StringIO()
+    for n, code in enumerate(env["_ih"]):
+        if code:
+            source.write("# %%%% In [%s]:\n" % n)
+            source.write(code)
+            source.write("\n\n")
     return source.getvalue()
 
 
+def get_ipython_cells():
+    # type: () -> List[Tuple[Optional[str], Optional[Any]]]
+    """
+    Get the Notebook (ipython, colab, Jupyter, PyCharm, etc.) from the
+    history. Assumes that this method is run in an ipython environment.
+
+    Returns the command-history [(input,output), ...]
+
+    """
+    import IPython
+
+    ipy = IPython.get_ipython()
+    env = ipy.ns_table["user_local"]
+    cells = []
+    for n, code in enumerate(env["_ih"]):
+        output = env["_oh"].get(n, None)
+        if code or output:
+            cells.append((code, output))
+    return cells
+
+
+def _format_text(text):
+    # type: (Any) -> List[str]
+    """
+    Format multi-line text in the ipython style.
+    """
+    if not isinstance(text, six.string_types):
+        text = repr(text)
+
+    lines = text.splitlines(True)  # type: List[str]
+    # Remove potentially trailing newlines in the final element
+    lines[-1] = lines[-1].rstrip("\r\n").rstrip("\n")
+    return lines
+
+
+def _format_output(output, execution_count):
+    # type: (Any, int) -> Optional[Dict]
+    """
+    Get the mime bundle, etc. for each output form.
+    """
+    if output is None:
+        return None
+
+    if hasattr(output, "_repr_mimebundle_"):
+        return {
+            "data": output._repr_mimebundle_(),
+            "execution_count": execution_count,
+            "metadata": {},
+            "output_type": "execute_result",
+        }
+    else:
+        return {
+            "data": {"text/plain": _format_text(output)},
+            "execution_count": execution_count,
+            "metadata": {},
+            "output_type": "execute_result",
+        }
+
+
+def create_notebook(cells):
+    # type: (List[Tuple[Optional[str], Optional[Any]]]) -> Dict
+    """
+    From a list of (input, output) tuples, create a JSON notebook.
+    """
+    final_cells = []
+    for n, cell in enumerate(cells):
+        new_cell = {
+            "cell_type": "code",
+            "execution_count": n + 1,
+            "metadata": {},
+            "source": _format_text(cell[0]),
+        }
+
+        if cell[1] is not None:
+            new_cell["output_type"] = "execute_result"
+            new_cell["outputs"] = [_format_output(cell[1], n + 1)]
+        else:
+            new_cell["outputs"] = []
+
+        final_cells.append(new_cell)
+
+    return {"nbformat": 4, "nbformat_minor": 0, "metadata": {}, "cells": final_cells}
+
+
+def get_ipython_notebook():
+    # type: () -> Dict
+    return create_notebook(get_ipython_cells())
+
+
 def postprocess_gcp_cloud_metadata(cloud_metadata):
     # type: (Dict[str, Any]) -> Dict[str, Any]
 
@@ -231,22 +318,12 @@ CLOUD_METADATA_MAPPING = {
     },
 }
 
-ProviderMapping = TypedDict(
-    "ProviderMapping",
-    {
-        "url": str,
-        "headers": Dict[str, str],
-        "postprocess_function": Optional[Callable[[Dict[str, Any]], Dict[str, Any]]],
-    },
-)
-CloudDetails = TypedDict("CloudDetails", {"provider": str, "metadata": Dict[str, Any]})
-
 
 def get_env_cloud_details(timeout=1):
-    # type: (int) -> Optional[CloudDetails]
+    # type: (int) -> Optional[Any]
     for provider in CLOUD_METADATA_MAPPING.keys():
         try:
-            params = cast(ProviderMapping, CLOUD_METADATA_MAPPING[provider])
+            params = CLOUD_METADATA_MAPPING[provider]
             response = requests.get(
                 params["url"], headers=params["headers"], timeout=timeout
             )
