@@ -38,6 +38,7 @@ from ._typing import (
     UserText,
     ValidFilePath,
 )
+from .config import get_config
 from .connection import Reporting, get_http_session
 from .exceptions import FileIsTooBig
 from .logging_messages import (
@@ -82,12 +83,14 @@ def _send_file(url, files, params, experiment_id):
     # type: (str, Any, Dict[str, str], str) -> Response
     LOGGER.debug("Uploading files %r to %s with params %s", files, url, params)
 
+    config = get_config()
+
     with get_http_session() as session:
         r = session.post(
             url,
             params=params,
             files=files,
-            timeout=180,
+            timeout=config["comet.timeout.file_upload"],
             headers={"X-COMET-DEBUG-EXPERIMENT-KEY": experiment_id},
         )  # type: Response
 
@@ -329,18 +332,21 @@ class BaseUploadProcessor(object):
 
     def __init__(
         self,
-        user_input,
-        upload_limit,
-        url_params,
-        metadata,
-        copy_to_tmp,
-        error_message_identifier,
+        user_input,  # type: Any
+        upload_limit,  # type: int
+        url_params,  # type: Dict[str, Optional[Any]]
+        metadata,  # type: Optional[Dict[str, str]]
+        copy_to_tmp,  # type: bool
+        error_message_identifier,  # type: Any
+        tmp_dir,  # type: str
     ):
+        # type: (...) -> None
         self.user_input = user_input
         self.url_params = url_params
         self.metadata = self._validate_metadata(metadata)
         self.upload_limit = upload_limit
         self.error_message_identifier = error_message_identifier
+        self.tmp_dir = tmp_dir
 
         self.copy_to_tmp = copy_to_tmp
 
@@ -459,7 +465,7 @@ class BaseUploadProcessor(object):
         if self.copy_to_tmp:
             LOGGER.debug("Saving IO to tmp_file because of copy_to_tmp")
             # Convert the file-like to a temporary file on disk
-            file_path = write_file_like_to_tmp_file(io_object)
+            file_path = write_file_like_to_tmp_file(io_object, self.tmp_dir)
             self.copy_to_tmp = False
 
             # TODO it would be easier to use the same field name for a file or a figure upload
@@ -492,7 +498,7 @@ class BaseUploadProcessor(object):
                 # We couldn't convert to an io_object
                 return None
 
-            file_path = write_file_like_to_tmp_file(io_object)
+            file_path = write_file_like_to_tmp_file(io_object, self.tmp_dir)
 
             return self._process_upload_by_filepath(TemporaryFilePath(file_path))
 
@@ -529,14 +535,16 @@ class AssetUploadProcessor(BaseUploadProcessor):
 
     def __init__(
         self,
-        user_input,
-        upload_type,
-        url_params,
-        metadata,
-        upload_limit,
-        copy_to_tmp,
-        error_message_identifier,
+        user_input,  # type: Any
+        upload_type,  # type: str
+        url_params,  # type: Dict[str, Optional[Any]]
+        metadata,  # type: Optional[Dict[str, str]]
+        upload_limit,  # type: int
+        copy_to_tmp,  # type: bool
+        error_message_identifier,  # type: Any
+        tmp_dir,  # type: str
     ):
+        # type: (...) -> None
         self.UPLOAD_TYPE = upload_type
 
         if metadata is None:
@@ -551,6 +559,7 @@ class AssetUploadProcessor(BaseUploadProcessor):
             metadata,
             copy_to_tmp,
             error_message_identifier,
+            tmp_dir,
         )
 
     def process_upload_by_filepath(self, upload_filepath):
@@ -601,21 +610,23 @@ class ImageUploadProcessor(BaseUploadProcessor):
 
     def __init__(
         self,
-        user_input,
-        name,
-        overwrite,
+        user_input,  # type: Any
+        name,  # type: Optional[str]
+        overwrite,  # type: bool
         image_format,
         image_scale,
         image_shape,
         image_colormap,
         image_minmax,
         image_channels,
-        upload_limit,
-        url_params,
-        metadata,
-        copy_to_tmp,
-        error_message_identifier,
+        upload_limit,  # type: int
+        url_params,  # type: Dict[str, Optional[Any]]
+        metadata,  # type: Optional[Dict[str, str]]
+        copy_to_tmp,  # type: bool
+        error_message_identifier,  # type: Any
+        tmp_dir,
     ):
+        # type: (...) -> None
         self.name = name
         self.image_format = image_format
         self.image_scale = image_scale
@@ -630,6 +641,7 @@ class ImageUploadProcessor(BaseUploadProcessor):
             metadata,
             copy_to_tmp,
             error_message_identifier,
+            tmp_dir,
         )
 
     def process_upload_by_filepath(self, upload_filepath):
@@ -690,15 +702,17 @@ class AudioUploadProcessor(BaseUploadProcessor):
 
     def __init__(
         self,
-        user_input,
-        sample_rate,
-        overwrite,
-        upload_limit,
-        url_params,
-        metadata,
-        copy_to_tmp,
-        error_message_identifier,
+        user_input,  # type: Any
+        sample_rate,  # type: Optional[int]
+        overwrite,  # type: bool
+        upload_limit,  # type: int
+        url_params,  # type: Dict[str, Optional[Any]]
+        metadata,  # type: Optional[Dict[str, str]]
+        copy_to_tmp,  # type: bool
+        error_message_identifier,  # type: Any
+        tmp_dir,  # type: str
     ):
+        # type: (...) -> None
         self.sample_rate = sample_rate
 
         if metadata is None:
@@ -713,6 +727,7 @@ class AudioUploadProcessor(BaseUploadProcessor):
             metadata,
             copy_to_tmp,
             error_message_identifier,
+            tmp_dir,
         )
 
     def process_upload_by_filepath(self, upload_filepath):
@@ -780,14 +795,16 @@ class AssetDataUploadProcessor(BaseUploadProcessor):
 
     def __init__(
         self,
-        user_input,
-        upload_type,
-        url_params,
-        metadata,
-        upload_limit,
-        copy_to_tmp,
-        error_message_identifier,
+        user_input,  # type: Any
+        upload_type,  # type: str
+        url_params,  # type: Dict[str, Optional[Any]]
+        metadata,  # type: Optional[Dict[str, str]]
+        upload_limit,  # type: int
+        copy_to_tmp,  # type: bool
+        error_message_identifier,  # type: Any
+        tmp_dir,  # type: str
     ):
+        # type: (...) -> None
         self.UPLOAD_TYPE = upload_type
         super(AssetDataUploadProcessor, self).__init__(
             user_input,
@@ -796,6 +813,7 @@ class AssetDataUploadProcessor(BaseUploadProcessor):
             metadata,
             copy_to_tmp,
             error_message_identifier,
+            tmp_dir,
         )
 
     def process_upload_to_be_converted(self, user_input):
