@@ -29,14 +29,14 @@ def xgboost_train(experiment, original, return_value, *args, **kwargs):
 
         # Log XGBoost eval metrics
         for key in kwargs["evals_result"]:
-            context = experiment.context
-            experiment.context = key
-            for metric_name in kwargs["evals_result"][key]:
-                for step, value in enumerate(kwargs["evals_result"][key][metric_name]):
-                    experiment._log_metric(
-                        metric_name, value, step=step, framework="xgboost"
-                    )
-            experiment.context = context
+            with experiment.context_manager(key):
+                for metric_name in kwargs["evals_result"][key]:
+                    for step, value in enumerate(
+                        kwargs["evals_result"][key][metric_name]
+                    ):
+                        experiment._log_metric(
+                            metric_name, value, step=step, framework="xgboost"
+                        )
 
 
 def patch(module_finder):
