@@ -18,7 +18,7 @@ MSG_FORMAT = "COMET %(levelname)s: %(message)s"
 
 FILE_MSG_FORMAT = "[%(process)d-%(processName)s:%(thread)d] %(relativeCreated)d COMET %(levelname)s [%(filename)s:%(lineno)d]: %(message)s"
 
-GO_TO_DOCS_MSG = " \nFor more details, please refer to: https://www.comet.ml/docs/python-sdk/warnings-errors"
+GO_TO_DOCS_MSG = " \nFor more details, please refer to: https://www.comet.ml/docs/python-sdk/warnings-errors/"
 
 INTERNET_CONNECTION_ERROR = (
     "Failed to establish connection to Comet server. Please check your internet connection. "
