@@ -96,7 +96,7 @@ class OfflineExperiment(BaseExperiment):
         api_key=None,  # type: Optional[str]
         display_summary=None,  # type: Optional[bool]
         log_env_cpu=True,  # type: Optional[bool]
-        display_summary_level=1,  # type: Optional[int]
+        display_summary_level=None,  # type: Optional[int]
     ):
         # type: (...) -> None
         """
