@@ -567,19 +567,27 @@ class Optimizer(object):
                 mode = "local"
             else:
                 mode = self.predictor_spec["mode"]
-            exp.set_predictor(
-                Predictor(
-                    exp,
-                    loss_name=self.predictor_spec["loss_name"],
-                    patience=self.predictor_spec["patience"],
-                    threshold=self.predictor_spec["threshold"],
-                    api=self.predictor_spec["api"],
-                    optimizer_id=self.id,
-                    interval=self.predictor_spec["interval"],
-                    start=self.predictor_spec["start"],
-                    mode=mode,
+
+            try:
+                exp.set_predictor(
+                    Predictor(
+                        exp,
+                        loss_name=self.predictor_spec["loss_name"],
+                        patience=self.predictor_spec["patience"],
+                        threshold=self.predictor_spec["threshold"],
+                        api=self.predictor_spec["api"],
+                        optimizer_id=self.id,
+                        interval=self.predictor_spec["interval"],
+                        start=self.predictor_spec["start"],
+                        mode=mode,
+                    )
                 )
-            )
+            except Exception:
+                LOGGER.warning(
+                    "Failure to create Predictor, experiment will run without Predictor",
+                    exc_info=True,
+                )
+
         exp.log_parameters(parameters)
         # Log optimizer static information:
         exp.log_other("optimizer_id", self.id)
