@@ -15,9 +15,14 @@
 import logging
 import time
 
-from ._typing import Any, Dict
+from ._typing import Any, Dict, List, Optional, Tuple
 from .config import get_config
-from .connection import LowLevelHTTPClient, get_optimizer_api_client, url_join
+from .connection import (
+    LowLevelHTTPClient,
+    get_optimizer_api_client,
+    sanitize_url,
+    url_join,
+)
 
 LOGGER = logging.getLogger(__name__)
 
@@ -30,15 +35,16 @@ class Predictor(object):
     def __init__(
         self,
         experiment,
+        mode=None,
         loss_name="loss",
         patience=10,
         best_callback=None,
         threshold=0.1,
         api=None,
         optimizer_id=None,
-        interval=2,
-        start=5,
-        mode=None,
+        start_step=10,
+        time_interval=300,
+        n_init_experiments=0,
         **defaults
     ):
         """
@@ -56,7 +62,6 @@ class Predictor(object):
         self.defaults = {
             "experiment_key": self.experiment.id,
             "api_key": self.experiment.api_key,
-            "TS": [],
             "HP_samples": float("nan"),
             "AP_no_parameters": float("nan"),
             "HP_epochs": float("nan"),
@@ -64,40 +69,47 @@ class Predictor(object):
             "HP_batch_size": float("nan"),
             "HP_curr_step": float("nan"),
         }
+        self.losses = []  # type: List[float]
         self.set_defaults(**defaults)
         self.threshold = threshold
         if not 0.0 < self.threshold <= 1.0:
             raise ValueError("Threshold must be set between 0.0 and 1.0")
 
-        self.interval = interval
-        # If interval is set to 0
-        if not self.interval >= 1:
-            raise ValueError("Please set interval to at least 1")
+        # The minimum number of data points from a loss curve
+        # required to make a prediction
+        self.min_samples = 10
 
-        self.predictor_url = self.config["comet.predictor_url"]
+        self.n_init_experiments = n_init_experiments
+        self.start_step = start_step
 
-        # Ensure we add a trailing slash, we want to append to the configured path
-        if not self.predictor_url.endswith("/"):
-            self.predictor_url += "/"
+        # The timestamp of the last call to the predictor server
+        self.last_prediction_timestamp = 0.0
+
+        self.stop_step = None
+        self.stop_val = None
+        self.stop_min = None
+
+        self.time_interval = time_interval
+        self.predictor_url = sanitize_url(self.config["comet.predictor_url"])
 
         self.predictor_ping_url = url_join(self.predictor_url, "isAlive/ping")
         self.predictor_predict_url = url_join(
             self.predictor_url, "lc_predictor/predict/"
         )
         self._low_level_http_client = LowLevelHTTPClient(
-            self.predictor_url, self.config["comet.timeout.http"]
+            self.predictor_url, self.config["comet.timeout.predictor"]
         )
 
         try:
             status = self.status()
+            self.experiment.log_other("predictor_id", status["model_id"])
         except Exception:
             LOGGER.error(
-                "Failed to contact Predictor at %s; aborting" % self.predictor_url
+                "Failed to contact Predictor at %s; ignoring" % self.predictor_url,
+                exc_info=True,
             )
-            raise
 
         self.experiment.log_other("predictor_loss_name", self.loss_name)
-        self.experiment.log_other("predictor_id", status["model_id"])
 
         self.MAX_TRIES = 5  # TODO: how many get_optimizer_best() retries are needed
 
@@ -118,8 +130,6 @@ class Predictor(object):
 
         elif mode == "global":
             self._setup_global_mode(optimizer_id)
-            self.start = start
-
             self.api = api if api else get_optimizer_api_client(self.experiment.api_key)
 
         elif mode == "local":
@@ -159,7 +169,7 @@ class Predictor(object):
             self.optimizer_id = optimizer_id
             self.experiment.log_other("optimizer_id", self.optimizer_id)
             LOGGER.info(
-                "Using optimizer_id %r explicitely given with Global mode",
+                "Using optimizer_id %r explicitly given with Global mode",
                 self.optimizer_id,
             )
 
@@ -169,9 +179,9 @@ class Predictor(object):
         """
         Please email lcp-beta@comet.ml for comments or questions.
         """
-        self.step = None
+        self.step = -1
         self.wait = 0
-        self.defaults["TS"] = []
+        self.losses = []
 
     def set_defaults(self, **defaults):
         """
@@ -194,163 +204,159 @@ class Predictor(object):
         Please email lcp-beta@comet.ml for comments or questions.
         """
         try:
-            loss = float(loss)
-        except Exception:
-            raise ValueError("Predictor.report_loss() requires a single number")
+            try:
+                loss = float(loss)
+            except Exception:
+                LOGGER.error(
+                    "Predictor.report_loss() requires a single number; ignoring %r",
+                    loss,
+                )
+                return
 
-        self.step = step if step is not None else self.experiment.curr_step
-        self.defaults["TS"].append(loss)
-        self.experiment.log_metric("predictor_tracked_loss", loss, step=self.step)
+            self.step = step if step is not None else self.experiment.curr_step
+            self.losses.append(loss)
+            self.experiment.log_metric("predictor_tracked_loss", loss, step=self.step)
+
+        except Exception:
+            LOGGER.error("Error reporting loss", exc_info=True)
+
+    def _elapsed_time(self):
+        return time.time() - self.last_prediction_timestamp
 
     def _local_stop_early(self, **data):
-        if len(self.defaults["TS"]) < 10:
+        # type: (Any) -> bool
+        current_loss = self.losses[-1]
+        lowest_min = min(self.losses[:-1])
+
+        current_best = min(lowest_min, current_loss)
+
+        if self.step is None:
+            step = float("nan")
+        else:
+            step = self.step
+
+        data.update({"best_metric": current_best, "HP_curr_step": step})
+        lmup = self.get_prediction(data)
+
+        if lmup is None:
             return False
 
-        if (len(self.defaults["TS"]) % self.interval) == 0:
-            current_loss = self.defaults["TS"][-1]
-            lowest_min = min(self.defaults["TS"][:-1])
+        (lower, mean, upper, p_improvement) = lmup
+        self.experiment.log_metrics(
+            {
+                "predictor_mean": mean,
+                "predictor_upper": upper,
+                "predictor_lower": lower,
+                "predictor_p_improvement": p_improvement,
+                "predictor_threshold": self.threshold,
+                "predictor_patience": self.patience,
+                "predictor_wait": self.wait,
+            },
+            step=self.step,
+        )
 
-            current_best = min(lowest_min, current_loss)
+        # If the loss is improving, reset the wait count
+        # Every time we see an improvement, run the callback
+        if current_loss < lowest_min:
+            self.wait = 0
 
-            if self.step is None:
-                step = float("nan")
-            else:
-                step = self.step
+        if p_improvement <= self.threshold:
+            self.wait += 1
 
-            data.update({"best_metric": current_best, "HP_curr_step": step})
-            lmup = self.get_prediction(**data)
+        # If patience is exceeded, stop training
+        if self.wait >= self.patience:
+            if self.stop_step is None:
+                self.stop_step = self.step
+                self.stop_val = self.losses[-1]
+                self.stop_min = min(self.losses)
 
-            if lmup is None:
-                return False
-
-            (lower, mean, upper, p_improvement) = lmup
-            self.experiment.log_metrics(
-                {
-                    "predictor_mean": mean,
-                    "predictor_upper": upper,
-                    "predictor_lower": lower,
-                    "predictor_p_improvement": p_improvement,
-                    "predictor_threshold": self.threshold,
-                    "predictor_patience": self.patience,
-                    "predictor_wait": self.wait,
-                },
-                step=self.step,
+            self.experiment.log_other("predictor_stop_step", self.stop_step)
+            msg = (
+                "Predicted probability of improvement failed"
+                " to exceed threshold {} for {} evaluations".format(
+                    self.threshold, self.patience
+                )
             )
+            self.stop_reason = msg
+            self.experiment.log_other("predictor_stop_reason", msg)
 
-            # If the loss is improving, reset the wait count
-            # Every time we see an improvement, run the callback
-            if current_loss < lowest_min:
-                self.wait = 0
+            if self.best_callback:
+                try:
+                    self.best_callback(self, current_best)
+                except Exception:
+                    LOGGER.error(
+                        "Error calling best callback with best value: %r",
+                        current_best,
+                        exc_info=True,
+                    )
 
-                if callable(self.best_callback):
-                    self.best_callback(self, current_loss)
-
-            # Else increment the wait count
-            else:
-                self.wait += 1
-
-            if p_improvement <= self.threshold:
-                msg = (
-                    "Predicted probability of improvement {:.3f}"
-                    " based on the best value {:.3f} seen in this trial for metric: {} "
-                    "is lower than the threshold {} "
-                )
-                self.experiment.log_other(
-                    "predictor_stop_step", self.experiment.curr_step
-                )
-                self.experiment.log_other(
-                    "predictor_stop_reason",
-                    msg.format(
-                        p_improvement, current_best, self.loss_name, self.threshold
-                    ),
-                )
-
-                return True
-
-            # If patience is exceeded, stop training
-            if self.wait >= self.patience:
-                self.experiment.log_other(
-                    "predictor_stop_step", self.experiment.curr_step
-                )
-                self.experiment.log_other(
-                    "predictor_stop_reason",
-                    "Patience value for this experiment has been exceeded",
-                )
-
-                return True
+            return True
 
         return False
 
     def _global_stop_early(self, **data):
-        # For cases where model is not stopped after _global_stop_early returns True
-        if self.wait >= self.patience:
-            LOGGER.debug("Patience Exceeded: %s >= %s", self.wait, self.patience)
-            return True
+        # type: (Any) -> bool
+        state = self._get_trial_state(data)
 
-        if len(self.defaults["TS"]) < 10:
+        if state is None:
             return False
 
-        if (len(self.defaults["TS"]) % self.interval) == 0:
-            state = self._get_trial_state(**data)
+        experiment_count = state.get("experiment_count")
+        best_metric = state.get("best_metric")
 
-            if state is None:
-                return False
+        lower = state.get("lower")
+        mean = state.get("mean")
+        upper = state.get("upper")
 
-            experiment_count = state.get("experiment_count")
-            best_metric = state.get("best_metric")
+        p_improvement = state.get("prob_improvement")
+        self.experiment.log_metrics(
+            {
+                "predictor_mean": mean,
+                "predictor_max": upper,
+                "predictor_min": lower,
+                "predictor_threshold": self.threshold,
+                "predictor_patience": self.patience,
+                "predictor_wait": self.wait,
+                "predictor_p_improvement": p_improvement,
+            }
+        )
 
-            lower = state.get("lower")
-            mean = state.get("mean")
-            upper = state.get("upper")
+        if experiment_count > self.n_init_experiments:
+            if p_improvement <= self.threshold:
+                self.wait += 1
 
-            p_improvement = state.get("prob_improvement")
-            self.experiment.log_metrics(
-                {
-                    "predictor_mean": mean,
-                    "predictor_max": upper,
-                    "predictor_min": lower,
-                    "predictor_threshold": self.threshold,
-                    "predictor_patience": self.patience,
-                    "predictor_wait": self.wait,
-                    "predictor_p_improvement": p_improvement,
-                }
-            )
+        if self.wait >= self.patience:
+            if self.stop_step is None:
+                self.stop_step = self.step
+                self.stop_val = self.losses[-1]
+                self.stop_min = min(self.losses)
 
-            # TODO: Move block into its own function
-            if experiment_count > self.start:
-                if p_improvement <= self.threshold:
-                    self.wait += 1
+                self.experiment.log_other("predictor_stop_step", self.stop_step)
+                msg = (
+                    "Predicted probability of improvement failed"
+                    " to exceed threshold {} for {} evaluations".format(
+                        self.threshold, self.patience
+                    )
+                )
+                self.experiment.log_other("predictor_stop_reason", msg)
+                self.stop_reason = msg
+                if self.best_callback:
+                    try:
+                        self.best_callback(self, best_metric)
 
-                    if self.wait >= self.patience:
-                        self.experiment.log_other(
-                            "predictor_stop_step", self.experiment.curr_step
-                        )
-                        msg = (
-                            "Predicted probability of improvement {:.3f}"
-                            " based on the best value {:.3f} seen in this trial for metric: {} "
-                            "is lower than the threshold {} "
-                        )
-                        self.experiment.log_other(
-                            "predictor_stop_reason",
-                            msg.format(
-                                p_improvement,
-                                best_metric,
-                                self.loss_name,
-                                self.threshold,
-                            ),
+                    except Exception:
+                        LOGGER.error(
+                            "Error calling best callback with best value: %r",
+                            best_metric,
+                            exc_info=True,
                         )
 
-                        return True
+                return True
 
         return False
 
-    def _get_trial_state(self, **data):
-        """
-        Internal method only used in global mode to retrieve best metric
-        in an optimizer run
-        :param data:
-        :return:
-        """
+    def _get_trial_metrics(self):
+        # type: () -> Optional[Tuple[Optional[int], Optional[float]]]
         count = 0
         while count <= self.MAX_TRIES:
             try:
@@ -358,16 +364,14 @@ class Predictor(object):
                 response_data = self.api.get_optimizer_best(
                     self.experiment.id, self.optimizer_id, self.loss_name, maximum=False
                 )
-            except KeyboardInterrupt:
-                raise
-
             except Exception as e:
                 if count == self.MAX_TRIES:
                     LOGGER.error(
                         "Max retries exceeded while trying to fetch "
-                        "experiments with current optimizer id"
+                        "experiments with current optimizer id",
+                        exc_info=True,
                     )
-                    raise
+                    return None
 
                 LOGGER.debug(
                     "Error while getting the optimizer best %s", e, exc_info=True
@@ -381,43 +385,109 @@ class Predictor(object):
         experiment_count = response_data.get("experimentCount")
         best_metric = response_data.get("metricValue")
 
+        return experiment_count, best_metric
+
+    def _get_trial_state(self, data):
+        # type: (Dict[str, Any]) -> Optional[Dict[str, float]]
+        """
+        Internal method only used in global mode to retrieve best metric
+        in an optimizer run
+        :param data:
+        :return:
+        """
+        trial_metrics = self._get_trial_metrics()
+
+        if trial_metrics is None:
+            return None
+
+        experiment_count, best_metric = trial_metrics
         if (experiment_count is None) or (best_metric is None):
-            return
+            return None
 
         data.update({"best_metric": best_metric, "HP_curr_step": self.step})
-        lmup = self.get_prediction(**data)
+        lmup = self.get_prediction(data)
 
-        if lmup:
-            lower, mean, upper, p_improvement = lmup
+        if lmup is None:
+            return None
 
-            return {
-                "lower": lower,
-                "mean": mean,
-                "upper": upper,
-                "prob_improvement": p_improvement,
-                "experiment_count": experiment_count,
-                "best_metric": best_metric,
-            }
-
-        return
+        lower, mean, upper, p_improvement = lmup
+        return {
+            "lower": lower,
+            "mean": mean,
+            "upper": upper,
+            "prob_improvement": p_improvement,
+            "experiment_count": experiment_count,
+            "best_metric": best_metric,
+        }
 
     def stop_early(self, **data):
+        # type: (Any) -> bool
         """
         Please email lcp-beta@comet.ml for comments or questions.
         """
-        if self.mode == "global":
-            return self._global_stop_early(**data)
+        # These checks should be performed in the order. Changing the order will
+        # produce unwanted behaviour
 
-        else:
-            return self._local_stop_early(**data)
+        try:
+            # Check for cases where model is not stopped after _global_stop_early returns True
+            if self.wait >= self.patience:
+                LOGGER.debug(
+                    "Patience Exceeded: %s >= %s; stopping", self.wait, self.patience
+                )
+                return True
 
-    def get_prediction(self, **request_data):
+            # If the step has not been set yet or the Experiment curr_step is None
+            if self.step is None:
+                LOGGER.debug("We don't have a step; not stopping")
+                return False
+
+            # Check if we have passed the start_step
+            if self.step < self.start_step:
+                LOGGER.debug(
+                    "Step %r < configured started step %r; not stopping",
+                    self.step,
+                    self.start_step,
+                )
+                return False
+
+            #  Check if we have the min number of samples necessary to make a prediction
+            if len(self.losses) < self.min_samples:
+                LOGGER.debug(
+                    "Number of epochs %r < configured min_samples %r; not stopping",
+                    len(self.losses),
+                    self.min_samples,
+                )
+                return False
+
+            # Check if enough time has passed between predictor evaluations
+            if self._elapsed_time() < self.time_interval:
+                LOGGER.debug(
+                    "Elapsed time since last call %r < configured time interval %r; not stopping",
+                    self._elapsed_time,
+                    self.time_interval,
+                )
+                return False
+
+            if self.mode == "global":
+                return self._global_stop_early(**data)
+            else:
+                return self._local_stop_early(**data)
+
+        except Exception:
+            LOGGER.error(
+                "Error calling Predictor.stop_early, not stopping", exc_info=True
+            )
+            return False
+
+    def get_prediction(self, request_data):
+        # type: (Dict[str, Any]) -> Optional[Tuple[float, float, float, float]]
         """
         Please email lcp-beta@comet.ml for comments or questions.
         """
 
         # Copy default fields
         prediction_data = self.defaults.copy()
+        prediction_data["TS"] = self.losses
 
         # Update fields
         prediction_data.update(request_data)
@@ -430,13 +500,13 @@ class Predictor(object):
         )
 
         response = self._low_level_http_client.post(
-            self.predictor_predict_url,
-            payload=prediction_request,
-            timeout=300,
-            retry=False,
+            self.predictor_predict_url, payload=prediction_request, retry=False
         )
 
         if response.status_code == 200:
+            # Update timestamp to latest successful call to predictor server
+            self.last_prediction_timestamp = time.time()
+
             response_data = response.json().get("response")
             self._set_prediction(response_data)
             return (
@@ -457,3 +527,24 @@ class Predictor(object):
 
     def _set_prediction(self, prediction):
         self.latest_prediction.update(prediction)
+
+    def get_notification_message(self):
+        msg = {
+            "stop-min": self.stop_min,
+            "stop-val": self.stop_val,
+            "stop-step": self.stop_step if self.stop_step else self.step,
+            "predicted-min": self.latest_prediction.get("mean"),
+            "predicted-min-lower_bound": self.latest_prediction.get("min"),
+            "predicted-min-upper_bound": self.latest_prediction.get("max"),
+        }
+
+        if self.mode == "global":
+            trial_metrics = self._get_trial_metrics()
+
+            if trial_metrics is not None:
+                _, best_metric = trial_metrics
+                msg["best-sweep-value"] = best_metric
+
+            return msg
+
+        return msg
