@@ -26,6 +26,7 @@ import operator
 import os
 import os.path
 import random
+import re
 import tempfile
 import time
 from collections import defaultdict
@@ -35,7 +36,20 @@ import six
 from pkg_resources import DistributionNotFound, get_distribution
 from requests.models import PreparedRequest
 
-from ._typing import IO, Any, Dict, Generator, List, Optional, Set, Tuple, Type
+from ._typing import (
+    IO,
+    Any,
+    Callable,
+    Dict,
+    Generator,
+    List,
+    Optional,
+    Sequence,
+    Set,
+    Tuple,
+    Type,
+    Union,
+)
 
 LOGGER = logging.getLogger(__name__)
 LOG_ONCE_CACHE = set()  # type: Set[str]
@@ -210,7 +224,7 @@ def image_data_to_file_like_object(
     image_minmax,
     image_channels,
 ):
-    # type: (Any, Optional[str], str, float, Optional[Tuple[int]], Optional[str], Optional[Tuple[float]], str) -> Optional[IO[bytes]]
+    # type: (Union[IO[bytes], Any], Optional[str], str, float, Optional[Sequence[int]], Optional[str], Optional[Sequence[float]], str) -> Union[IO[bytes], None, Any]
     """
     Ensure that the given image_data is converted to a file_like_object ready
     to be uploaded
@@ -300,10 +314,30 @@ def array_to_image_fp(
     image_minmax,
     image_channels,
 ):
+    # type: (Any, str, float, Optional[Sequence[int]], Optional[str], Optional[Sequence[float]], str) -> Optional[IO[bytes]]
     """
     Convert a numpy array to an in-memory image
     file pointer.
     """
+    image = array_to_image(
+        array, image_scale, image_shape, image_colormap, image_minmax, image_channels
+    )
+    return image_to_fp(image, image_format)
+
+
+def array_to_image(
+    array,
+    image_scale,
+    image_shape,
+    image_colormap,
+    image_minmax,
+    image_channels,
+    mode=None,
+):
+    # type: (Any, float, Optional[Sequence[int]], Optional[str], Optional[Sequence[float]], Optional[str], Optional[str]) -> Optional[Any]
+    """
+    Convert a numpy array to an in-memory image.
+    """
     try:
         import PIL.Image
         import numpy
@@ -312,19 +346,25 @@ def array_to_image_fp(
         LOGGER.error(
             "The Python libraries PIL, numpy, and matplotlib are required for this operation"
         )
-        return
+        return None
+
+    array = numpy.array(array)
 
     ## Handle image transformations here
     ## End up with a 0-255 PIL Image
     if image_minmax is not None:
         minmax = image_minmax
     else:  # auto minmax
-        minmax = [array.min(), array.max()]
-        if minmax[0] == minmax[1]:
-            minmax[0] = minmax[0] - 0.5
-            minmax[1] = minmax[1] + 0.5
-        minmax[0] = math.floor(minmax[0])
-        minmax[1] = math.ceil(minmax[1])
+        flatten_array = flatten(array)
+        min_array = min(flatten_array)
+        max_array = max(flatten_array)
+        if min_array == max_array:
+            min_array = min_array - 0.5
+            max_array = max_array + 0.5
+        min_array = math.floor(min_array)
+        max_array = math.ceil(max_array)
+        minmax = [min_array, max_array]
+
     ## if a shape is given, try to reshape it:
     if image_shape is not None:
         try:
@@ -332,6 +372,7 @@ def array_to_image_fp(
             array = array.reshape(image_shape[1], image_shape[0])
         except Exception:
             LOGGER.info("WARNING: invalid image_shape; ignored", exc_info=True)
+
     ## If 3D, but last array is flat, make it 2D:
     if len(array.shape) == 3 and array.shape[-1] == 1:
         array = array.reshape((array.shape[0], array.shape[1]))
@@ -340,6 +381,7 @@ def array_to_image_fp(
         array = numpy.array([array])
     if image_channels == "first" and len(array.shape) == 3:
         array = numpy.moveaxis(array, 0, -1)
+
     ### Ok, now let's colorize and scale
     if image_colormap is not None:
         ## Need to be in range (0,1) for colormapping:
@@ -357,17 +399,144 @@ def array_to_image_fp(
         ## Rescale (0, 255)
         array = rescale_array(array, minmax, (0, 255), "uint8")
         image = PIL.Image.fromarray(array)
+
     if image_scale != 1.0:
         image = image.resize(
             (int(image.size[0] * image_scale), int(image.size[1] * image_scale))
         )
+
     ## Put in a standard mode:
-    if image.mode not in ["RGB", "RGBA"]:
+    if mode:
+        image = image.convert(mode)
+    elif image.mode not in ["RGB", "RGBA"]:
         image = image.convert("RGB")
-    return image_to_fp(image, image_format)
+    return image
+
+
+def dataset_to_sprite_image(
+    matrix,
+    size,
+    preprocess_function=None,
+    transparent_color=None,
+    background_color_function=None,
+):
+    # type: (Any, Sequence[int], Optional[Callable], Optional[Sequence[int]], Optional[Callable]) -> Any
+    """
+    Convert a dataset (array of arrays) into a giant image of
+    images (a sprite sheet).
+
+    Args:
+        matrix: array of vectors or Images
+        size: (width, height) of each thumbnail image
+        preprocess_function: function to preprocess image values
+        transparent_color: color to mark as transparent
+        background_color_function: function that takes index, returns a color
+
+    Returns: image
+    """
+    try:
+        from PIL import Image
+    except ImportError:
+        LOGGER.error("The Python library PIL is required for this operation")
+        return None
+
+    length = len(matrix)
+    sprite_size = math.ceil(math.sqrt(length))
+
+    sprite_image = Image.new(
+        mode="RGBA",
+        size=(int(sprite_size * size[0]), int(sprite_size * size[1])),
+        color=(0, 0, 0, 0),
+    )
+    if preprocess_function is not None:
+        matrix = preprocess_function(matrix)
+    for i, array in enumerate(matrix):
+        if isinstance(array, Image.Image):
+            image = array
+        else:
+            image = array_to_image(
+                array,
+                image_scale=1.0,
+                image_shape=size,
+                image_colormap=None,
+                image_minmax=(0, 255),
+                image_channels="last",
+                mode="RGBA",
+            )
+
+            if not image:
+                return None
+
+        if transparent_color is not None:
+            image = image_transparent_color(image, transparent_color, threshold=1)
+        if background_color_function is not None:
+            color = background_color_function(i)
+            image = image_background_color(image, color)
+        if image.mode != "RGBA":
+            image = image.convert("RGBA")
+        location = (int((i % sprite_size) * size[0]), int((i // sprite_size) * size[1]))
+        sprite_image.paste(image, location)
+    return sprite_image
+
+
+def image_background_color(image, color):
+    # type: (Any, Sequence[int]) -> Optional[Any]
+    """
+    Given an image with some transparency, add a background color to it.
+
+    Args:
+        image: a PIL image
+        color: a red, green, blue color tuple
+    """
+    try:
+        from PIL import Image
+    except ImportError:
+        LOGGER.error("The Python library PIL is required for this operation")
+        return None
+
+    if image.mode != "RGBA":
+        raise ValueError(
+            "image must have an alpha channel in order to set a background color"
+        )
+
+    new_image = Image.new("RGB", image.size, color)
+    new_image.paste(image, (0, 0), image)
+    return new_image
+
+
+def image_transparent_color(image, color, threshold=1):
+    # type: (Any, Tuple[int, int, int], int) -> Any
+    """
+    Given a color, make that be the transparent color.
+
+    Args:
+        image: a PIL image
+        color: a red, green, blue color tuple
+        threshold: the max difference in each color component
+    """
+    try:
+        from PIL import Image
+        import numpy
+    except ImportError:
+        LOGGER.error(
+            "The Python libraries PIL and numpy are required for this operation"
+        )
+        return
+
+    image = image.convert("RGBA")
+    array = numpy.array(numpy.asarray(image))
+    r, g, b, a = numpy.rollaxis(array, axis=-1)
+    mask = (
+        (numpy.abs(r - color[0]) < threshold)
+        & (numpy.abs(g - color[1]) < threshold)
+        & (numpy.abs(b - color[2]) < threshold)
+    )
+    array[mask, 3] = 0
+    return Image.fromarray(array, mode="RGBA")
 
 
 def image_to_fp(image, image_format):
+    # type: (Any, str) -> IO[bytes]
     """
     Convert a PIL.Image into an in-memory file
     pointer.
@@ -427,8 +596,48 @@ def write_file_like_to_tmp_file(file_like_object, tmpdir):
     return tmp_file.name
 
 
+def table_to_fp(tabular_data, delim, headers):
+    # type: (Any, str, Union(Sequence[str], bool)) -> IO
+    if isinstance(headers, bool):
+        add_headers = headers
+    else:
+        add_headers = True
+    fp = StringIO()
+    for row in tabular_data:
+        if isinstance(row, (numbers.Number, six.string_types)):
+            if add_headers:
+                if isinstance(headers, bool):
+                    fp.write('"column1"\n')
+                else:
+                    fp.write('"%s"\n' % (headers[0],))
+                add_headers = False
+            fp.write(str(row) + "\n")
+        else:
+            columns = flatten(row)
+            if add_headers:
+                for i, column in enumerate(columns):
+                    if isinstance(headers, bool):
+                        column = '"column%s"' % (i + 1)
+                    else:
+                        column = '"%s"' % (headers[i],)
+                    if i == len(columns) - 1:
+                        fp.write(column)
+                    else:
+                        fp.write(column + delim)
+                fp.write("\n")
+                add_headers = False
+            for i, column in enumerate(columns):
+                if i == len(columns) - 1:
+                    fp.write(str(column))
+                else:
+                    fp.write(str(column) + delim)
+            fp.write("\n")
+    fp.seek(0)
+    return fp
+
+
 def data_to_fp(data):
-    # type: (Any) -> Optional[IO]
+    # type: (Union[bytes, str, Any]) -> Optional[IO]
     if isinstance(data, str):
         fp = StringIO()
         fp.write(data)
@@ -464,6 +673,49 @@ def format_bytes(size):
         return "%s %s" % (size, "bytes")
 
 
+class Embedding(object):
+    """
+    Data structure for holding embedding template info.
+    """
+
+    def __init__(
+        self,
+        vector_url,
+        vector_shape,
+        metadata_url,
+        sprite_url=None,
+        image_size=None,
+        title="Comet Embedding",
+    ):
+        """
+        The URL's and sizes for all of the Embedding'
+        resources.
+        """
+        self.vector_url = vector_url
+        self.vector_shape = vector_shape
+        self.metadata_url = metadata_url
+        self.sprite_url = sprite_url
+        self.image_size = image_size
+        self.title = title
+
+    def to_json(self):
+        """
+        Return a JSON representation of the embedding
+        """
+        template = {
+            "tensorName": self.title,
+            "tensorShape": list(self.vector_shape),
+            "tensorPath": self.vector_url,
+            "metadataPath": self.metadata_url,
+        }
+        if self.sprite_url is not None:
+            template["sprite"] = {
+                "imagePath": self.sprite_url,
+                "singleImageDim": list(self.image_size),
+            }
+        return template
+
+
 class ConfusionMatrix(object):
     """
     Data structure for holding a confusion matrix of values and their
@@ -584,10 +836,10 @@ class ConfusionMatrix(object):
             raise ValueError("if you give y_true OR y_predicted you must give both")
 
         if winner_function is not None:
-            self.winner_function = winner_function
+            self.winner_function = winner_function  # type: ignore
 
         if index_to_example_function is not None:
-            self.index_to_example_function = index_to_example_function
+            self.index_to_example_function = index_to_example_function  # type: ignore
 
         self.y_true = y_true
         self.y_predicted = y_predicted
@@ -1109,16 +1361,7 @@ class Histogram(object):
         if len(values) == 0:
             return
 
-        # Numpy arrays have an optimized method of flattening:
-        if hasattr(values, "flatten"):
-            values = values.flatten()
-        else:
-            # Otherwise, we try to flatten via functools.reduce:
-            try:
-                values = functools.reduce(operator.iconcat, values, [])
-            except TypeError:
-                # Otherwise, assume that it is already flat:
-                pass
+        values = flatten(values)
 
         # Sort for speed of inserts
         if counts is None:
@@ -1510,7 +1753,7 @@ def get_file_extension(file_path):
 
 
 def encode_metadata(metadata):
-    # type: (Optional[Dict[Any, Any]]) -> Optional[str]
+    # type: (Optional[Dict[Any, Any]]) -> Optional[Union[str, bytes]]
     if metadata is None:
         return None
 
@@ -1665,6 +1908,41 @@ def convert_to_matrix(matrix, dtype=None):
         return [convert_to_list(row, dtype=dtype) for row in matrix]
 
 
+def shape(data):
+    """
+    Given a nested list or a numpy array,
+    return the shape.
+    """
+    if hasattr(data, "shape"):
+        return list(data.shape)
+    else:
+        try:
+            length = len(data)
+            return [length] + shape(data[0])
+        except TypeError:
+            return []
+
+
+def flatten(values):
+    """
+    Given a nested list or a numpy array,
+    return the data flattened.
+    """
+    # Numpy arrays have an optimized method of flattening:
+    if hasattr(values, "flatten"):
+        values = values.flatten()
+    else:
+        # Otherwise, we try to recursively flatten via functools.reduce:
+        try:
+            values = functools.reduce(
+                operator.iconcat, [flatten(value) for value in values], []
+            )
+        except TypeError:
+            # Otherwise, assume that it is already flat:
+            pass
+    return values
+
+
 def convert_to_list(items, dtype=None):
     """
     Take an unknown item and convert to a list of scalars
@@ -1800,3 +2078,16 @@ def proper_registry_model_name(name):
     while "--" in name:
         name = name.replace("--", "-")
     return name
+
+
+def safe_filename(filename):
+    """
+    Given a value, turn it into a valid filename.
+
+    1. Remove the spaces
+    2. Replace anything not alpha, '-', '_', or '.' with '_'
+    3. Remove duplicate '_'
+    """
+    string = str(filename).strip().replace(" ", "_")
+    string = re.sub(r"(?u)[^-\w.]", "_", string)
+    return re.sub(r"_+", "_", string)
